/**
 * 
 * File: FuzzyKNN.java
 * 
 * The Fuzzy K-NN algorithm. 
 * 
 * @author Written by Joaqun Derrac (University of Granada) 21/06/2011 
 * @version 1.0 
 * @since JDK1.5
 * 
 */

package keel.Algorithms.FingerprintClassification.Jain99;

import java.util.Arrays;
import java.util.StringTokenizer;

import org.core.Files;
import org.core.Randomize;

import keel.Algorithms.FingerprintClassification.FingerprintClassificationAlgorithm;
import keel.Algorithms.FingerprintClassification.ReportTool;
import keel.Algorithms.FingerprintClassification.Timer;
import keel.Algorithms.FingerprintClassification.Util;

public class Jain99 extends FingerprintClassificationAlgorithm {

	Jain99NN networks [][];
	private int k;
	private int hidden;
	private double learning,momentum;
	
	private int trainingMask [];
	
	private int firstClass,secondClass;
	private int votes1,votes2;
	
	private static final double VALIDATION = 0.3;
	/** 
	 * Reads the parameters of the algorithm. 
	 * 
	 * @param script Configuration script
	 * 
	 */
	@Override
	protected void readParameters(String script) {
		
		String file;
		String line;
		StringTokenizer fileLines, tokens;
		
	    file = Files.readFile (script);
	    fileLines = new StringTokenizer (file,"\n\r");
	    
	    //Discard in/out files definition
	    fileLines.nextToken();
	    fileLines.nextToken();
	    fileLines.nextToken();
	    
	    //Getting the seed
	    line = fileLines.nextToken();
	    tokens = new StringTokenizer (line, "=");
	    tokens.nextToken();
	    seed = Long.parseLong(tokens.nextToken().substring(1));
	    
	    //Getting the K parameter
	    line = fileLines.nextToken();
	    tokens = new StringTokenizer (line, "=");
	    tokens.nextToken();
	    k = Integer.parseInt(tokens.nextToken().substring(1));
	    
	    //Getting the number of hidden neurons
	    line = fileLines.nextToken();
	    tokens = new StringTokenizer (line, "=");
	    tokens.nextToken();
	    hidden = Integer.parseInt(tokens.nextToken().substring(1));
	    
	    //Getting the learning rate
	    line = fileLines.nextToken();
	    tokens = new StringTokenizer (line, "=");
	    tokens.nextToken();
	    learning = Double.parseDouble(tokens.nextToken().substring(1));
	    
	    //Getting the learning rate
	    line = fileLines.nextToken();
	    tokens = new StringTokenizer (line, "=");
	    tokens.nextToken();
	    momentum = Double.parseDouble(tokens.nextToken().substring(1));    
	    
	} //end-method	

	/**
	 * Main builder. Initializes the methods' structures
	 * 
	 * @param script Configuration script
	 */
	public Jain99(String script){
		
		readDataFiles(script);
		
		//Naming the algorithm
		name="Jain99";
		
		networks= new Jain99NN[nClasses][nClasses];
		
		trainingMask = new int [trainData.length];

		//Initialization of random generator    
	    Randomize.setSeed(seed);

	    //Initialization of Reporting tool
	    ReportTool.setOutputFile(outFile[2]);
	    
	} //end-method	
	
	/**
	 * Generates the model of the algorithm
	 */
	public void generateModel (){
		
		int nTrain,nVal;
		double error,errorEpoch,errorValidation, oldErrorValidation;
		double improvement;
		
		//Start of model time
		Timer.resetTime();	
		
		//create the OVO neural networks
		
		for(int i=0;i<nClasses;i++){
			for(int j=i+1;j<nClasses;j++){
				networks[i][j]=new Jain99NN(inputAtt,2,hidden,learning,momentum);
			}
		}
		
		//train the OVO neural networks
		for(int i=0;i<nClasses;i++){
			for(int j=i+1;j<nClasses;j++){
				System.out.println("Training network "+i+" VS "+j);
				
				//mark instances as unusable=0, training=1 and validation=2
				
				nTrain=0;
				nVal=0;
				Arrays.fill(trainingMask,0);
				
				for(int k=0;k<trainData.length;k++){
					
					if((trainOutput[k]==i)||(trainOutput[k]==j)){
						//VALIDATION (i.e. 30%) instances are randomly selected for validation
						if(Randomize.Rand()<=VALIDATION){
							trainingMask[k]=2;
							nVal++;
						}
						else{
							trainingMask[k]=1;
							nTrain++;
						}
							
					}
				}
				
				improvement= 1.0;
				oldErrorValidation=Double.MAX_VALUE;
				
				int min=1000;int max=10000; int current=0;
				while(((current<min)||(improvement>0))&&(current<max)){
					current++;
					//training process (one epoch)
					errorEpoch=0;
					for(int k=0;k<trainData.length;k++){
							
						if(trainingMask[k]==1){
							
							if(trainOutput[k]==i){
								error=networks[i][j].train(trainData[k], 0);
							}
							else{
								error=networks[i][j].train(trainData[k], 1);
							}
							
							errorEpoch+=error;
						}	
							
						
					}
				
					errorValidation=0;
					
					//validation process (one epoch)
					for(int k=0;k<trainData.length;k++){
						
						if(trainingMask[k]==2){
							
							if(trainOutput[k]==i){
								error=networks[i][j].validate(trainData[k], 0);
							}
							else{
								error=networks[i][j].validate(trainData[k], 1);
							}
							
							errorValidation+=error;
						}		
						
					}
								
					System.out.println("Validation error: "+errorValidation);
					
					improvement=oldErrorValidation-errorValidation;
					oldErrorValidation=errorValidation;	
				}
			}
		}

		//End of model time
		Timer.setModelTime();
	
		//Showing results
		System.out.println(name+" "+ relation + " Model " + Timer.getModelTime() + "s");
		
	}
	
	
	/**
	 * Classifies the training set (leave-one-out)
	 */
	public void classifyTrain(){
	    
		//Start of training time
		Timer.resetTime();
		
		classifyTrainSet();
		
		//End of training time
		Timer.setTrainingTime();
		
		//Showing results
		System.out.println(name+" "+ relation + " Training " + Timer.getTrainingTime() + "s");
		
	} //end-method	
	
	/**
	 * Classifies the test set
	 */
	public void classifyTest(){
		    
		//Start of training time
		Timer.resetTime();
		
		classifyTestSet();
		
		//End of test time
		Timer.setTestTime();
		
		//Showing results
		System.out.println(name+" "+ relation + " Test " + Timer.getTestTime() + "s");	
		
	} //end-method	
	
	/**
	 * Classifies the training set
	 */
	public void classifyTrainSet(){
				
		for(int i=0;i<trainData.length;i++){
			
			trainPrediction[i]=classifyInstance(i,true);
			
		}

	} //end-method	
	
	/**
	 * Classifies the test set
	 */
	public void classifyTestSet(){

		for(int i=0;i<testData.length;i++){
			
			testPrediction[i]=classifyInstance(i,false);
			
		}

	} //end-method	
	
	private int classifyInstance(int index, boolean training){
		
		int output=-1;
		
		//kNN classification
		knnClassify(index,training);
		
		output=firstClass;
		
		if(votes2==0){
			return output;
		}
		
		//OVO NN classification
		if(training){
			if(firstClass<secondClass){
				output=networks[firstClass][secondClass].classify(trainData[index]);
				if(output==0){
					output=firstClass;
				}
				else{
					output=secondClass;
				}
			}
			else{
				output=networks[secondClass][firstClass].classify(trainData[index]);
				if(output==0){
					output=secondClass;
				}
				else{
					output=firstClass;
				}
			}
		}
		else{
			System.out.println("NEW: "+firstClass+" "+secondClass);
			if(firstClass<secondClass){
				output=networks[firstClass][secondClass].classify(testData[index]);
				if(output==0){
					output=firstClass;
				}
				else{
					output=secondClass;
				}
			}
			else{
				output=networks[secondClass][firstClass].classify(testData[index]);
				if(output==0){
					output=secondClass;
				}
				else{
					output=firstClass;
				}
			}
		}
		
		return output;
	}
	
	private void knnClassify(int index, boolean training){
		
		double minDist[];
		int nearestN[];
		double dist;
		boolean stop;
		int selectedClasses [];

		nearestN = new int[k];
		minDist = new double[k];
	
	    Arrays.fill(nearestN, -1);
	    Arrays.fill(minDist, Double.MAX_VALUE);
		
	    //KNN Method starts here
	    
		for (int i=0; i<trainData.length; i++) {
		
			if((!training)||(i!=index)){ //leave-one-out
				
				if(training){
					dist = Util.euclideanDistance(trainData[i],trainData[index]);
				}else{
					dist = Util.euclideanDistance(trainData[i],testData[index]);
				}
			    
	
				//see if it's nearer than our previous selected neighbors
				stop=false;
					
				for(int j=0;j<k && !stop;j++){
					
					if (dist < minDist[j]) {
						    
						for (int l = k - 1; l >= j+1; l--) {
							minDist[l] = minDist[l - 1];
							nearestN[l] = nearestN[l - 1];
						}	
							
						minDist[j] = dist;
						nearestN[j] = i;
						stop=true;
					}
				}
			}
		}
		
		
		selectedClasses= new int[nClasses];
			
		Arrays.fill(selectedClasses, 0);	
				
		for (int i=0; i<k; i++) {
			selectedClasses[trainOutput[nearestN[i]]]+=1;
		}
				
		if(selectedClasses[0]>=selectedClasses[1]){
			firstClass=0;
			secondClass=1;
			votes1=selectedClasses[0];
			votes2=selectedClasses[1];
		}
		else{
			firstClass=1;
			secondClass=0;
			votes1=selectedClasses[1];
			votes2=selectedClasses[0];
			
		}
				
		for (int i=2; i<nClasses; i++) {
			if (votes1 < selectedClasses[i]) {
			
				votes2=votes1;
				votes1=selectedClasses[i];
				
				secondClass=firstClass;
				firstClass=i;
				
			}
			else{
				if (votes2 < selectedClasses[i]) {
					
					votes2=selectedClasses[i];
					
					secondClass=i;
				
				}
			}
		}

	}
	
	
	/**
	 * Reports the results obtained
	 */
	public void printReport(){
		
		writeOutput(outFile[0], trainOutput, trainPrediction);
		writeOutput(outFile[1], testOutput, testPrediction);
		
		ReportTool.setResults(trainOutput,trainPrediction,testOutput,testPrediction,nClasses);
		
		ReportTool.printReport();
		
	}
} //end-class 
