/**
 * 
 * File: FuzzyKNN.java
 * 
 * The Fuzzy K-NN algorithm. 
 * 
 * @author Written by Joaqun Derrac (University of Granada) 21/06/2011 
 * @version 1.0 
 * @since JDK1.5
 * 
 */

package keel.Algorithms.FingerprintClassification.Leung;

import java.util.Arrays;
import java.util.StringTokenizer;

import org.core.Files;

import keel.Algorithms.FingerprintClassification.FingerprintClassificationAlgorithm;
import keel.Algorithms.FingerprintClassification.ReportTool;
import keel.Algorithms.FingerprintClassification.Timer;

import Jama.*;

public class Leung extends FingerprintClassificationAlgorithm {

	int elemClass [];
	private double meanVectors [][];
	private Matrix SW;
	private Matrix SB;
	
	private EigenvalueDecomposition decomposition;
	private Matrix blockDiagonal;
	private Matrix eigenVectors;
	
	private int dimensions;
	private Matrix newDataM;
	private double newData [][];
	private int newInputAtt;
	
	private int regularizationEigenValue;
	private double regularizationValue;

	private double QDAregularizationValue;
	
	private Matrix covariancesM [];
	private Matrix QDAmeansM;
	private double QDAmeans[][];
	private double QDAcovariances [][][];
	
	private EigenvalueDecomposition CVdecompositions [];
	private Matrix CVeigenVectors [];
	private double CVeigenValues [][];
	
	private double QDAConstantTerms [];
	
	
	/** 
	 * Reads the parameters of the algorithm. 
	 * 
	 * @param script Configuration script
	 * 
	 */
	@Override
	protected void readParameters(String script) {
		
		String file;
		String line;
		String type;
		StringTokenizer fileLines, tokens;
		
	    file = Files.readFile (script);
	    fileLines = new StringTokenizer (file,"\n\r");
	    
	    //Discard in/out files definition
	    fileLines.nextToken();
	    fileLines.nextToken();
	    fileLines.nextToken();
	    
	    //Getting the number of dimensions
	    line = fileLines.nextToken();
	    tokens = new StringTokenizer (line, "=");
	    tokens.nextToken();
	    dimensions = Integer.parseInt(tokens.nextToken().substring(1));
	    
	    //Getting the regularizationEigenValue
	    line = fileLines.nextToken();
	    tokens = new StringTokenizer (line, "=");
	    tokens.nextToken();
	    regularizationEigenValue = Integer.parseInt(tokens.nextToken().substring(1));
	    
	    //Getting a fixed regularization value
	    line = fileLines.nextToken();
	    tokens = new StringTokenizer (line, "=");
	    tokens.nextToken();
	    regularizationValue = Double.parseDouble(tokens.nextToken().substring(1));
	    
	    //Getting a fixed regularization value
	    line = fileLines.nextToken();
	    tokens = new StringTokenizer (line, "=");
	    tokens.nextToken();
	    QDAregularizationValue = Double.parseDouble(tokens.nextToken().substring(1));
	    
	   
	    
	} //end-method	

	/**
	 * Main builder. Initializes the methods' structures
	 * 
	 * @param script Configuration script
	 */
	public Leung(String script){
		
		readDataFiles(script);
		
		//Naming the algorithm
		name="Leung";
		
		meanVectors = new double [nClasses][inputAtt];
		elemClass = new int [nClasses];

	    //Initialization of Reporting tool
	    ReportTool.setOutputFile(outFile[2]);
	    
	} //end-method	
	
	/**
	 * Generates the model of the algorithm
	 */
	public void generateModel (){
		
		//Start of model time
		Timer.resetTime();	
		
		//Step 1: Modified Fisher LDA
		computeMeans();
		
		computeSMatrixes();
		
		computeEigenValues();
		
		reduceTrainingSet();
		
		//Step 2: Modified QDA
		
		computeQDAStatistics();

		//End of model time
		Timer.setModelTime();
	
		//Showing results
		System.out.println(name+" "+ relation + " Model " + Timer.getModelTime() + "s");
		
	}
	
	private void computeQDAStatistics(){
		
		int classAt;
		double factor;
		
		//compute means
		QDAmeans = new double [nClasses][newInputAtt];
		
		for(int k=0;k<nClasses;k++){
			Arrays.fill(QDAmeans[k], 0.0);
		}
		
		for(int i=0;i<newData.length;i++){
			classAt=trainOutput[i];
			for(int j=0;j<newInputAtt;j++){
				QDAmeans[classAt][j]+=newData[i][j];
			}
		}
		
		for(int k=0;k<nClasses;k++){
			factor=(double)elemClass[k];
			for(int j=0;j<newInputAtt;j++){
				QDAmeans[k][j]/=factor;
			}
			
		}
		
		QDAmeansM= new Matrix(QDAmeans);
		
		//compute covariances
		QDAcovariances = new double[nClasses][newInputAtt][newInputAtt];
		
		for(int k=0;k<nClasses;k++){
			for(int j=0;j<newInputAtt;j++){
			 Arrays.fill(QDAcovariances[k][j], 0.0);
			}
		}
		
		for (int j = 0; j < newInputAtt; j++) {
				
			for (int j2 = j; j2 < newInputAtt; j2++) {
					
				for(int i=0;i<newData.length;i++){
					
					classAt=trainOutput[i];
					QDAcovariances[classAt][j][j2] += ((newData[i][j]-QDAmeans[classAt][j]) * (newData[i][j2]-QDAmeans[classAt][j2]));
			
				}
			}
		}
		
		for(int k=0;k<nClasses;k++){
			for (int j = 0; j < newInputAtt; j++) {	
				for (int j2 = j; j2 < newInputAtt; j2++) {
					QDAcovariances[k][j][j2] = QDAcovariances[k][j][j2]	/ (elemClass[k]-1);
					QDAcovariances[k][j2][j] = QDAcovariances[k][j][j2];
				}
			}
		}
		
		//generate matrixes
		covariancesM= new Matrix[nClasses];
		for(int k=0;k<nClasses;k++){
			covariancesM[k]=new Matrix(QDAcovariances[k]);
		}
		
		//obtain covariance eigenvalue decompositions
		CVdecompositions = new EigenvalueDecomposition[nClasses];
		CVeigenValues = new double [nClasses][];
		CVeigenVectors = new Matrix [nClasses];
		
		for(int k=0;k<nClasses;k++){
			CVdecompositions[k]=new EigenvalueDecomposition(covariancesM[k]);
			CVeigenValues[k]=CVdecompositions[k].getRealEigenvalues();
			CVeigenVectors[k]=CVdecompositions[k].getV();
		}
		
		//compute QDA class constants term
		
		 QDAConstantTerms= new double [nClasses];
		 
		 double prod;
		 
		 for(int k=0;k<nClasses;k++){
			 
			 prod=1.0;
			 
			 for (int j = 0; j < newInputAtt; j++) {
				 
				 prod*=CVeigenValues[k][j]+QDAregularizationValue;
			 }
			 
			 QDAConstantTerms[k]=Math.log(prod);
		 }
		
	}
	
	private void reduceTrainingSet(){
	
		Matrix factors=eigenVectors.getMatrix(0, eigenVectors.getRowDimension()-1, 0, dimensions-1);
		
		newDataM =  new Matrix (trainData);
		
		newDataM= newDataM.times(factors);
		
		newData= newDataM.getArrayCopy();
		
		newInputAtt=newData[0].length;
		
	}
	
	
	
	private void computeEigenValues(){
		
		Matrix invSW;
		Matrix fisherDisc;
		
		invSW=SW.inverse();
		
		fisherDisc=invSW.times(SB);
		
		decomposition = new EigenvalueDecomposition(fisherDisc);
		
		blockDiagonal=decomposition.getD();
		eigenVectors=decomposition.getV();
		
	}

	private void computeSMatrixes(){
		
		int classAt;
		
		Matrix a,b,c,d;
		Matrix e [];
		
		double regularization;
		
		//compute SW
		e= new Matrix[nClasses];
		for(int k=0;k<nClasses;k++){
			e[k]= new Matrix(inputAtt,inputAtt);
		}
		
		for(int i=0;i<trainData.length;i++){
			classAt=trainOutput[i];
			a=new Matrix(trainData[i],1);
			b=new Matrix(meanVectors[classAt],1);
			c=a.minus(b);
			d=(c.transpose()).times(c);
			e[classAt].plusEquals(d);
		}
		
		for(int k=0;k<nClasses;k++){	
			e[k]=e[k].times(1.0/(double)(elemClass[k]));
		}
		
		//e[k] now has Sum(class(k)) (x-mean)*(x-mean)'    /#elements 
		
		SW=e[0];
		
		for(int k=1;k<nClasses;k++){	
			SW=SW.plusEquals(e[k]);
		}
		
		//add regularization constant
		regularization = 0;
				
		if(regularizationEigenValue==-1){
			regularization=regularizationValue;
		}
		else{
			EigenvalueDecomposition eig = new EigenvalueDecomposition(SW);
			
			if(regularizationEigenValue<eig.getRealEigenvalues().length){
				regularization=eig.getRealEigenvalues()[regularizationEigenValue];
			}
			else{
				regularization = 0;
			}
		}
		
		Matrix add = Matrix.identity(SW.getRowDimension(), SW.getColumnDimension());
		
		add.times(regularization);
		
		SW=SW.plusEquals(add);
		
		//compute SB
		
		if(nClasses==2){
			a=new Matrix(meanVectors[0],1);
			b=new Matrix(meanVectors[1],1);
			c=a.minusEquals(b);
			SB=(c.transpose()).times(c);
		}
		else{
			//compute mean of class means
			double absoluteMean []=new double[inputAtt];
			Arrays.fill(absoluteMean, 0.0);
			
			for(int k=0;k<nClasses;k++){
				for(int j=0;j<inputAtt;j++){
					absoluteMean[j]+=meanVectors[k][j];
				}
			}
			
			for(int j=0;j<inputAtt;j++){
				absoluteMean[j]/=(double)nClasses;
			}
			
			//compute SB
			SB = new Matrix(inputAtt,inputAtt,0);
			for(int k=0;k<nClasses;k++){
				a=new Matrix(meanVectors[k],1);
				b=new Matrix(absoluteMean,1);
				c=a.minusEquals(b);
				d=(c.transpose()).times(c);
				SB=SB.plusEquals(d);
			}
			SB = SB.times(1.0/(double)nClasses);
		}
		
	}
	
	private void computeMeans(){
		
		int classAt;
		double factor;
		
		
		Arrays.fill(elemClass, 0);
		for(int k=0;k<nClasses;k++){
			Arrays.fill(meanVectors[k], 0.0);
		}
		
		for(int i=0;i<trainData.length;i++){
			classAt=trainOutput[i];
			for(int j=0;j<inputAtt;j++){
				meanVectors[classAt][j]+=trainData[i][j];
			}
			elemClass[classAt]++;
		}
		
		for(int k=0;k<nClasses;k++){
			factor=(double)elemClass[k];
			for(int j=0;j<inputAtt;j++){
				meanVectors[k][j]/=factor;
			}
		}

	}
	
	private int classifyInstance(double example[]){
		
		double discriminantValues [];
		int output;
		
		discriminantValues=computeDiscriminantFunctions(example);
			
		double max=discriminantValues[0];
		output=0;
		
		for(int k=1;k<nClasses;k++){
			if(max<discriminantValues[k]){
				max=discriminantValues[k];
				output=k;
			}
		}
		
		return output;
	}
	
	private double [] computeDiscriminantFunctions(double example[]){
		
		double values [] = new double [nClasses];
		double a,b,sum;
		Matrix tip,eigV;
		
		Arrays.fill(values, 0);
		
		//reduce example through Fisher's discriminants

		double exampleM [][] = new double [nClasses][example.length];
		
		for(int k=0;k<nClasses;k++){
			System.arraycopy(example, 0, exampleM[k], 0, example.length);
		}

		Matrix reduced= new Matrix(exampleM);
		Matrix factors=eigenVectors.getMatrix(0, eigenVectors.getRowDimension()-1, 0, dimensions-1);
		
		reduced= reduced.times(factors);
		
		//tipify (substract mean) example through each class
		Matrix tipified= reduced.minus(QDAmeansM);
		
		for(int k=0;k<nClasses;k++){
			
			sum=0;
			for(int j=0;j<dimensions;j++){
				
				a=1.0/(CVeigenValues[k][j]+QDAregularizationValue);
				
				
				tip=tipified.getMatrix(k,k,0,dimensions-1);
				
				tip=tip.transpose();
				
				eigV=CVeigenVectors[k].getMatrix(0,dimensions-1,j,j);
				
				eigV=eigV.transpose();
				
				b=eigV.times(tip).get(0, 0);
				
				b=b*b;
				
				sum-=a*b;
			}
			
			values[k]=sum-QDAConstantTerms[k];

		}
		
		
		return values;
		
	}
	
	/**
	 * Classifies the training set (leave-one-out)
	 */
	public void classifyTrain(){
	    
		//Start of training time
		Timer.resetTime();
		
		classifyTrainSet();
		
		//End of training time
		Timer.setTrainingTime();
		
		//Showing results
		System.out.println(name+" "+ relation + " Training " + Timer.getTrainingTime() + "s");
		
	} //end-method	
	
	/**
	 * Classifies the test set
	 */
	public void classifyTest(){
		    
		//Start of training time
		Timer.resetTime();
		
		classifyTestSet();
		
		//End of test time
		Timer.setTestTime();
		
		//Showing results
		System.out.println(name+" "+ relation + " Test " + Timer.getTestTime() + "s");	
		
	} //end-method	
	
	/**
	 * Classifies the training set
	 */
	public void classifyTrainSet(){
				
		for(int i=0;i<trainData.length;i++){
						
			trainPrediction[i]=classifyInstance(trainData[i]);
			
		}

	} //end-method	
	
	/**
	 * Classifies the test set
	 */
	public void classifyTestSet(){

		for(int i=0;i<testData.length;i++){
						
			testPrediction[i]=classifyInstance(testData[i]);
			
		}

	} //end-method	
	

	
	
	/**
	 * Reports the results obtained
	 */
	public void printReport(){
		
		writeOutput(outFile[0], trainOutput, trainPrediction);
		writeOutput(outFile[1], testOutput, testPrediction);
		
		ReportTool.setResults(trainOutput,trainPrediction,testOutput,testPrediction,nClasses);
		
		ReportTool.printReport();

		
	} //end-method	
    
} //end-class 
